/*******************************************************************************
Copyright © 2023 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

(function(window) {

function checkCode_onlyOneCommand(oldCommand, newCommand) {
    return function({
        tokens      // Quellcode als Array von Token-Instanzen
    }) {
        const realTokens = tokens.filter(token => (!(token instanceof TokenBracketOpen) && !(token instanceof TokenBracketClose) && !(token instanceof TokenBlockComment) && !(token instanceof TokenLineComment) && !(token instanceof TokenWhitespace )&& !(token instanceof TokenIndent)));
        let wrongTokens = realTokens.filter(token => !(token instanceof TokenCommand));
        if (wrongTokens.length)
            return { accept: false, msg: 'Schön, dass du schon mehr als nur Methodenaufrufe kannst. Bei dieser Aufgabe sollen aber nur Algorithmen getestet werden, die aus nur einem Methodenaufruf bestehen', line: wrongTokens[0].line };
        wrongTokens = realTokens.filter(token => (token.name == oldCommand));
        if (wrongTokens.length)
            return { accept: false, fine: 1, msg: 'Du sollst den neuen Methodenaufruf ANSTATT des bisherigen Methodenaufrufs schreiben', line: wrongTokens[0].line };
        if (!realTokens.length)
            return { accept: false, fine: 1, msg: 'Der Methodenaufruf fehlt' };
        if (realTokens.length > 1)
            return { accept: false, fine: 0, msg: 'Verwende in dieser Aufgabe NUR EINEN Methodenaufruf', line: realTokens[1].line };
        if (realTokens[0].name != newCommand)
            return { accept: false, fine: 1, msg: 'Statt der Methode „' + realTokens[0].name + '“ sollst du die Methode „' + newCommand + '“ ausprobieren', line: realTokens[0].line };
        return { accept: true };
    };
}

const LEVELS = [
    {
        // Angezeigter Arbeitsauftrag (HTML erlaubt)
        task: 'Ins Texteingabefeld wird der Algorithmus für den Roboter eingegeben. Aktuell kannst du ihn noch nicht verändern. Errate erst, was der aktuelle Algorithmus bewirkt!',

        // Welt und Roboterstartposition als Maschinencode
        world: '0533000000000100020000112',

        // Quelltext, der anfangs im Editor steht
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nvor()',

        // Ist das Bearbeiten des Quelltextes gesperrt? (Optional, Standardwert false)
        readonly: true,

        // Soll der Quellcode nicht geändert werden, wenn zuvor das direkt vorherige Level aktiv war?
        keepCode: false,

        // Muss - ggf. zusätzlich zur korrekten Fragenbeantwortung - checkCode oder checkTermination unlock=true liefern, bevor Level abgeschlossen werden kann? (Optional, Standardwert false)
        locked: false,

        // Optional: Methode, die beim Start des Levels ausgeführt werden soll
        start() { },

        // Optional: Überprüft, ob ein syntaktisch korrekter Quellcode ausgeführt werden darf; Rückgabewert true für ja, sonst { accept: "Ausführung erlauben?", line: "Zeilennummer eines Fehlers (optional)", msg: "Fehlermeldung (optional)", fine: "Als einen Fehlversuch werten? (optional)", unlock: "Level als abgeschlossen behandeln? (optional)" }
        checkCode({
            code,       // In Editor eingegebener Code
            tokens,     // Quellcode als Array von Token-Instanzen
            compiled,   // Quelltextbaum als SourceCodePart-Instanz
            worldDef    // Maschinencode der Welt zu Beginn
        }) {
            return { accept: false, msg: "Geduld! Ausprobieren darfst du das bei Aufgabe 2…" };
        },

        // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
        checkTermination({
            code,       // In Editor eingegebener Code
            line,       // Zuletzt ausgeführte Zeilennummer
            result,     // Grund für das Ende: undefined bei manuellem Abbruch, sonst RobotSimulator.RESULT_*
            worldDef,   // Maschinencode der Welt zu Beginn
            worldStat,  // Present robot's world after execution (RobotWorld instance)
            robotStat   // Location and viewing direction of robot after execution: { x, y, direction }
        }) {
            return {};
        },

        // Wird falsches Beantworten der Frage als Fehlversuch gewertet? (Optional, Standardwert false)
        questionWrongFine: false,

        // Falls erst eine Frage richtig beantwortet werden muss: Welche?
        question: {
            // Optional: Titel der Frage
            title: undefined,

            // Frage
            html: "Was bewirkt der Algorithmus wohl?",

            // Optional: Fragentext, nach Überprüfen der Antwort
            htmlSolution: "Was bewirkt der Algorithmus wohl?",

            // Optional: Kann nur eine Antwort richtig sein?
            singlechoice: false,

            // Optional: Sollen die Antworten nicht durchgemischt werden?
            keepOrder: false,

            // Optional: Zusätzliche Erläuterung, wenn Frage falsch beantwortet wird
            response: "Beachte, dass das grau Geschriebene in Zeile 1 ein Kommentar ist, der bei der Algorithmusausführung ignoriert wird",

            // Auswählbare Antworten
            choices: [
                {
                    // Antwort
                    html: "Der Roboter geht auf ein anderes Feld",

                    // Optional: Ist das eine richtige Antwort?
                    correct: true,

                    // Optional: Anzuzeigender Text bei Fehlauswahl
                    response: undefined
                },
                { html: "Bei <em>genau einem</em> Feld ändert sich die Anzahl der darauf liegenden Plättchen" },
                { html: "Auf <em>mehreren</em> Feldern ändert sich die Anzahl der Plättchen" },
                { html: "Der Roboter dreht sich" },
                { html: "Die Welt ändert ihre Größe", response: "Die Weltgröße und die Art der Felder kann bei Roboter! nicht durch den Algorithmus verändert werden." },
                { html: "Nichts" }
            ]
        }
    },
    {
        task: 'Drücke nun auf <code>Roboter starten</code>, um den Algorithmus auszuprobieren!',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nvor()',
        readonly: true,
        keepCode: true,
        locked: true,

        // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
        checkTermination() {
            return { unlock: true };
        }
    },
    {
        task: '<p>Beachte: Wenn du den Roboter erneut startest, wird seine Welt immer erst in den Ausgangszustand zurückgesetzt.</p><p>Probiere das aus, indem du den Algorithmus mindestens dreimal hintereinander startest: Jedes mal befindet sich der Roboter wieder an seiner ursprünglichen Startposition, anschließend geht er &dash; wie vom Algorithmus festgelegt &dash; ein Feld nach vorne.</p>',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nvor()',
        readonly: true,
        keepCode: true,
        locked: true,

        start() { this._countDone = 0; },

        // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
        checkTermination({ result }) {
            if (result != RobotSimulator.RESULT_DONE)
                return {};
            return { unlock: (++this._countDone >= 3) };
        }
    },
    {
        task: 'Nun kannst du den Algorithmus verändern: Finde nun heraus, was <code>plus</code> bewirkt!',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nvor()',
        keepCode: true,
        locked: true,
        checkCode: checkCode_onlyOneCommand('vor', 'plus'),

        // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
        checkTermination() {
            return { unlock: true };
        },
        questionWrongFine: 1,
        question: {
            html: "Was bewirkt ein Aufruf der Methode <code>plus</code>?",
            response: "Probiere erneut obigen Code aus, um die Wirkung des Methodenaufrufs zu beobachten!",

            // Auswählbare Antworten
            choices: [
                { html: "Der Roboter geht auf ein anderes Feld" },
                { html: "Bei <em>genau einem</em> Feld ändert sich die Anzahl der darauf liegenden Plättchen", correct: true },
                { html: "Auf <em>mehreren</em> Feldern ändert sich die Anzahl der Plättchen" },
                { html: "Der Roboter dreht sich" },
                { html: "Die Welt ändert ihre Größe", response: "Die Weltgröße und die Art der Felder kann bei Roboter! nicht durch den Algorithmus verändert werden." },
                { html: "Nichts" }
            ]
        }
    },
    {
        task: 'Finde wieder durch Ausprobieren heraus, was <code>minus</code> bewirkt!',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nplus()',
        keepCode: true,
        locked: true,
        checkCode: checkCode_onlyOneCommand('plus', 'minus'),

        // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
        checkTermination() {
            return { unlock: true };
        },
        questionWrongFine: 1,
        question: {
            html: "Was bewirkt ein Aufruf der Methode <code>minus</code>?",
            response: "Probiere erneut obigen Code aus, um die Wirkung des Methodenaufrufs zu beobachten!",

            // Auswählbare Antworten
            choices: [
                { html: "Der Roboter geht auf ein anderes Feld" },
                { html: "Bei <em>genau einem</em> Feld ändert sich die Anzahl der darauf liegenden Plättchen", correct: true },
                { html: "Auf <em>mehreren</em> Feldern ändert sich die Anzahl der Plättchen" },
                { html: "Der Roboter dreht sich" },
                { html: "Die Welt ändert ihre Größe", response: "Die Weltgröße und die Art der Felder kann bei Roboter! nicht durch den Algorithmus verändert werden." },
                { html: "Nichts" }
            ]
        }
    },
    {
        task: 'Finde durch Ausprobieren noch heraus, was <code>linksrum</code> bewirkt!',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nminus()',
        keepCode: true,
        locked: true,
        checkCode: checkCode_onlyOneCommand('minus', 'linksrum'),

        // Optional: Reagiert auf ein Ende der Algorithmusausführung; Rückgabewert wie checkCode
        checkTermination() {
            return { unlock: true };
        },
        questionWrongFine: 1,
        question: {
            html: "Was bewirkt ein Aufruf der Methode <code>linksrum</code>?",
            response: "Probiere erneut obigen Code aus, um die Wirkung des Methodenaufrufs zu beobachten!",

            // Auswählbare Antworten
            choices: [
                { html: "Der Roboter geht auf ein anderes Feld" },
                { html: "Bei <em>genau einem</em> Feld ändert sich die Anzahl der darauf liegenden Plättchen" },
                { html: "Auf <em>mehreren</em> Feldern ändert sich die Anzahl der Plättchen" },
                { html: "Der Roboter dreht sich", correct: true },
                { html: "Die Welt ändert ihre Größe", response: "Die Weltgröße und die Art der Felder kann bei Roboter! nicht durch den Algorithmus verändert werden." },
                { html: "Nichts" }
            ]
        }
    },
    {
        task: 'Methodenaufrufe innerhalb eines Algorithmus werden auch <strong>Anweisungen</strong> genannt.',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nlinksrum()',
        locked: false,
        readonly: true,
        questionWrongFine: 1,
        question: {
            html: "Wie viele <em>Anweisungen</em> enthält der obige Algorithmus?",
            response: "Nur Methodenaufrufe wie <code>vor()</code> zählen als Anweisungen",

            // Auswählbare Antworten
            choices: [
                { html: "Es gibt ", unit: " Anweisungen.", type: "number", accept: 1, min: 0, step: 1, correct: true, selectalways: true }
            ]
        }
    },
    {
        task: '<p>In dieser Programmiersprache sind Klammerpaare (also <code>()</code>) bei Anweisungen sogar optional, können also weggelassen werden.</p><p>Probiere jetzt aus, ob dieser Algorithmus genauso funktioniert, wenn du das Klammerpaar bei der Anweisung weg lässt!</p>',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nlinksrum()',
        keepCode: true,
        locked: true,

        checkCode({
            tokens     // Quellcode als Array von Token-Instanzen
        }) {
            const realTokens = tokens.filter(token => (!(token instanceof TokenBracketOpen) && !(token instanceof TokenBracketClose) && !(token instanceof TokenBlockComment) && !(token instanceof TokenLineComment) && !(token instanceof TokenWhitespace )&& !(token instanceof TokenIndent)));
            const wrongTokens = realTokens.filter(token => !(token instanceof TokenCommand));
            if (wrongTokens.length)
                return { accept: false, msg: 'Schön, dass du schon mehr als nur Anweisungen kannst. Bei dieser Aufgabe sollen aber nur Algorithmen getestet werden, die aus nur einer Anweisung bestehen', line: wrongTokens[0].line };
            if (!realTokens.length)
                return { accept: false, fine: 1, msg: 'Die Anweisung fehlt' };
            if (realTokens.length > 1)
                return { accept: false, fine: 0, msg: 'Verwende in dieser Aufgabe NUR EINE Anweisung', line: realTokens[1].line };
            if (realTokens[0].name != 'linksrum')
                return { accept: false, fine: 1, msg: 'Statt der Anweisung „' + realTokens[0].name + '“ sollst du die Anweisung „linksrum“ ausprobieren', line: realTokens[0].line };
            const bracketOpen = tokens.reduce((prev, curr) => (prev || ((curr instanceof TokenBracketOpen) ? curr : null)), null);
            if (bracketOpen)
                return { accept: false, msg: 'Warum hast du das Klammerpaar nicht entfernt?', line: bracketOpen.line, fine: 1 };
            return { accept: true };
        },

        checkTermination() { return { unlock: true }; }
    },
    {
        task: 'Alles ab einem Doppelkreuz <code>#</code> bis zum Zeilenende ist ein sogenannter <strong>(Zeilen-)Kommentar</strong>. Kommentare werden vom Roboter vollständig ignoriert, können aber nützlich sein, damit Menschen den Algorithmus schneller verstehen, oder für zusätzliche Anmerkungen.',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nlinksrum',
        locked: false,
        readonly: true,
        questionWrongFine: 1,
        question: {
            html: "Wie viele <em>Kommentare</em> enthält der obige Algorithmus?",

            // Auswählbare Antworten
            choices: [
                { html: "Es gibt ", unit: " Kommentare.", type: "number", accept: 3, min: 0, step: 1, correct: true, selectalways: true }
            ]
        }
    },
    {
        task: '<p>Übrigens kann ein Zeilenkommentar in dieser Programmiersprache statt mit <code>#</code> auch mit <code>;</code>, <code>//</code> oder <code>--</code> beginnen.</p><p>Probiere das Einfügen eines Zeilenkommentars aus, indem du einen Zeilenkommentar direkt hinter die Anweisung ergänzt!</p>',
        world: '0533000000000100020000112',
        code: '# Das grau Gedruckte wird vom Roboter ignoriert\n# Verfügbare Methoden: vor, plus, minus, linksrum\n\n# Ein einzelner Methodenaufruf geht so:\nlinksrum',
        keepCode: true,
        locked: true,

        checkCode({
            tokens      // Quellcode als Array von Token-Instanzen
        }) {
            const realTokens = tokens.filter(token => (!(token instanceof TokenBracketOpen) && !(token instanceof TokenBracketClose) && !(token instanceof TokenBlockComment) && !(token instanceof TokenLineComment) && !(token instanceof TokenWhitespace )&& !(token instanceof TokenIndent)));
            const wrongTokens = realTokens.filter(token => !(token instanceof TokenCommand));
            if (wrongTokens.length)
                return { accept: false, msg: 'Schön, dass du schon mehr als nur Anweisungen kannst. Bei dieser Aufgabe sollen aber nur Algorithmen getestet werden, die aus nur einer Anweisung bestehen', line: wrongTokens[0].line };
            if (!realTokens.length)
                return { accept: false, fine: 1, msg: 'Die Anweisung fehlt' };
            if (realTokens.length > 1)
                return { accept: false, fine: 0, msg: 'Verwende in dieser Aufgabe NUR EINE Anweisung', line: realTokens[1].line };
            if (realTokens[0].name != 'linksrum')
                return { accept: false, fine: 1, msg: 'Statt der Anweisung „' + realTokens[0].name + '“ sollst du die Anweisung „linksrum“ ausprobieren', line: realTokens[0].line };
            let foundCommand = false;
            for (let i = 0; i < tokens.length; ++i) {
                if (foundCommand && (tokens[i] instanceof TokenLineComment))
                    return { accept: true };
                if (foundCommand && (tokens[i] instanceof TokenBlockComment))
                    return { accept: false, msg: 'Gefragt ist hier ein Zeilenkommentar, kein Blockkommentar', line: foundCommand };
                if (foundCommand && (tokens[i] instanceof TokenIndent))
                    break;
                if (foundCommand)
                    continue;
                foundCommand = (tokens[i] instanceof TokenCommand) ? tokens[i].line : false;
            }
            return { accept: false, line: foundCommand, msg: 'Der Zeilenkommentar direkt hinter der Anweisung fehlt' };
        },

        checkTermination() { return { unlock: true }; }
    }
];

window.roboterCodingTask = {
    levels: LEVELS
};
})(window);
