/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/**
* Klasse für einfache Programmieraufgaben mit mehreren Teilaufgaben
*/
function RoboterCodingLmsConnector(
    node        // DOM-Knoten, der für die Benutzeroberfläche genutzt werden soll
) {
    this.upgradeLicenceBtn();
    this._ui = new RoboterCodingTaskUI(node);
    this.connect();
    this._ui.onlevelunlock = (() => this.saveScore());
    this._ui.onnewfine = (() => this.saveScore());
}

// Genutzte RoboterCodingTaskUI-Instanz
RoboterCodingLmsConnector.prototype._ui = null;

/**
* Schaltfläche zum Einblenden der Lizenz-Infos anzeigen
* @return undefined
*/
RoboterCodingLmsConnector.prototype.upgradeLicenceBtn = function() {
    new ScoTools.InfoPageLink();
}

/**
* Versuchen, eine Verbindung zur Lernplattform aufzubauen
* @return undefined
*/
RoboterCodingLmsConnector.prototype.connect = function() {
    if (typeof ScoTools.LMSLink == 'function') {
        var self = this;
        this.lmslink = new ScoTools.LMSLink();
        this.lmslink.onnew = (function() {
            self.lmsValuesInit();
        });
        this.lmslink.start();
        // Werte laden
        try {
            var saved = JSON.parse('[' + this.lmslink.lessonLocation + ']');
            if (saved.length >= 2)
                this._ui.loadValues(+saved[0], +saved[1]);
        } catch(e) {}
    }
    if (!this.lmslink.ready)
        this._ui.setGeneralMsg('Verbindungsaufnahme zur Lernplattform fehlgeschlagen - Fortschritte können nicht gespeichert werden', 'msg_warn');
}

/**
* Beim ersten Start des Lernpaketes Anfangswerte festlegen
* @return undefined
*/
RoboterCodingLmsConnector.prototype.lmsValuesInit = function() {
    this.lmslink.lessonLocation = '0,0';
}

// Genutzte Instanz von LMSLink
RoboterCodingLmsConnector.prototype.lmslink = { 'ready': false, lessonLocation: '0,0' };

/**
* Ermittelt, ob die Aufgabe bereits erfolgreich gelöst wurde
* @return Boolean       Aufgabe bereits gelöst?
*/
RoboterCodingLmsConnector.prototype.getDone = function() {
    return (this.lmslink.ready && !this.lmslink.isIncomplete());
}

/**
* Prüft, ob Veränderungen am Lernpaket gespeichert werden sollen
* @return Boolean       Kann und soll gespeichert werden?
*/
RoboterCodingLmsConnector.prototype.canSave = function() {
    return (!this.lmslink.ready || this.lmslink.isNormalIncomplete());
}

/**
* Bearbeitungszeit speichern
* @return undefined
*/
RoboterCodingLmsConnector.prototype.reportTime = function() {
    this.lmslink.ready && this.lmslink.reportTime();
}

/**
* Punktestand und Status speichern
* @return undefined
*/
RoboterCodingLmsConnector.prototype.saveScore = function(
    scoreRaw,       // Erreichter Punktestand
    scoreMin,       // Minimal erreichbarer Punktestand
    scoreMax,       // Maximal erreichbarer Punktestand
    status          // Optional: lessonStatus
) {
    if (!this.canSave()) {
        this.reportTime();
        return;
    }
    this.lmslink.lessonLocation = this._ui.unlockedLevelCount + ',' + this._ui.fineCount;
    if (this._ui.done)
        this.lmslink.saveScore(Math.min(100, Math.max(0, 100 - this._ui.fineCount)), 0, 100, "completed");
}

/**
* Einen Wert vom LMS lesen
* @return undefined
*/
RoboterCodingLmsConnector.prototype.read = function(
    key     // Attributname von LMSLink-Instanz, die gelesen werden soll
) {
    return (this.lmslink.ready ? this.lmslink[key] : undefined);
}

/**
* Einen Wert vom LMS speichern lassen
* @return undefined
*/
RoboterCodingLmsConnector.prototype.write = function(
    key,        // Attributname von LMSLink-Instanz, unter dem der Wert gespeichert werden soll
    value       // Zu speichernder Wert
) {
    if (this.lmslink.ready)
        this.lmslink[key] = value;
}

/**
* Zugleich mehrere Werte ändern, die Zeit speichern und dann erst Commit an das LMS senden
* @return undefined
*/
RoboterCodingLmsConnector.prototype.save = function(
    data        // Objekt mit zu speichernden Daten
) {
    if (this.lmslink.ready)
        this.lmslink.save(data);
}
