/**
* Zeigt durch roboterCodingTask definierte Level zum Lösen an
*/
function RoboterCodingTaskUI(
    nodeContainer       // DOM element to occupy
) {
    nodeContainer.innerHTML = '';
    this._levelMgr = new Levelliste(nodeContainer);
    this._levelMgr.onchange = ((levelNr, oldLevelNr) => this._startLevel(levelNr, oldLevelNr));
    this._levelMgr.strLevelName = 'Aufgabe ';
    this._levelMgr.onnamelevel = (nr => (nr > roboterCodingTask.levels.length) ? 'Abschluss' : '');
    this._levelMgr.count = roboterCodingTask.levels.length + 1;
    this._showTask('Wird geladen…');
    this._contentContainer = ScoTools.createHtml('div', nodeContainer);
    this._terminal = new RobotTerminal();
    this._contentContainer.appendChild(this._terminal.node);
    this._terminal.onbeforeExecution = ((code, tokens, compiled, worldDef) => this._procBeforeExecution(code, tokens, compiled, worldDef));
    this._terminal.onafterExecution = ((code, line, result, worldDef, worldStat, robotStat) => this._procAfterExecution(code, line, result, worldDef, worldStat, robotStat));
    this._levelMgr.active = 0;
}

// Verwendete Levelliste-Instanz
RoboterCodingTaskUI.prototype._levelMgr = null;

// Verwendete RobotTerminal-Instanz
RoboterCodingTaskUI.prototype._terminal = null;

// Container (DOM-Knoten) für Terminal, ggf. Testfrage und Weiter-Schaltfläche
RoboterCodingTaskUI.prototype._contentContainer = null;

// Container (DOM-Knoten) für Testfrage
RoboterCodingTaskUI.prototype._containerQuestion = null;

// DOM-Knoten für „Weiter“-Schaltfläche
RoboterCodingTaskUI.prototype._btnNext = null;

// Was muss als nächstes getan werden? (0 → Programmieraufgabe lösen, 1 → Frage richtig beantworten, 2 → Nichts mehr)
RoboterCodingTaskUI.prototype.unlockNext = 0;

// Funktion, die beim Freischalten eines neuen Levels aufgerufen werden soll
RoboterCodingTaskUI.prototype.onlevelunlock = function(unlockedLevelCount) { }

// Zahl der bereits freigeschalteten Level
Object.defineProperty(RoboterCodingTaskUI.prototype, 'unlockedLevelCount', {
    enumerable: true,
    get() { return Math.min(this._levelMgr.done, roboterCodingTask.levels.length); },
    set(v) {
        this._levelMgr.done = v + ((roboterCodingTask.levels.length == v) ? 1 : 0);
        this._levelMgr.active = v;
    }
});

// Wurden bereits alle Level frei geschaltet?
Object.defineProperty(RoboterCodingTaskUI.prototype, 'done', {
    enumerable: true,
    get() { return (this._levelMgr.done > roboterCodingTask.levels.length); }
});

// Funktion, die beim Erhalt von Strafpunkten aufgerufen werden soll
RoboterCodingTaskUI.prototype.onnewfine = function(totalFine, additionalFine) { }

// Gesammelte Anzahl an Fehlerpunkten
RoboterCodingTaskUI.prototype._fineCount = 0;
Object.defineProperty(RoboterCodingTaskUI.prototype, 'fineCount', {
    enumerable: true,
    get() { return this._fineCount; },
    set(v) {
        if ((v == this._fineCount) || !Number.isFinite(v))
            return;
        const oldFine = this._fineCount;
        this._fineCount = v;
        if ((typeof this.onnewfine == 'function') || (this.onnewfine instanceof Function))
            this.onnewfine.call(this, v, v - oldFine);
    }
});

/**
* Früher gespeicherte Werte wieder übernehmen
* @return undefined
*/
RoboterCodingTaskUI.prototype.loadValues = function(
    unlockedLevelCount,     // Anzahl der freigeschalteten Level
    fineCount               // Anzahl der gesammelten Fehlerpunkte
) {
    this._fineCount = (Number.isFinite(fineCount) ? fineCount : 0);
    this.unlockedLevelCount = unlockedLevelCount;
}

/**
* Aufgabenstellung anzeigen
* @return undefined
*/
RoboterCodingTaskUI.prototype._showTask = function(
    task        // Anzuzeigender HTML-Code
) {
    if (!task && this._nodeTask) {
        this._nodeTask.parentNode.removeChild(this._nodeTask);
        this._nodeTask = null;
        return;
    }
    if (this._nodeTask) {
        this._nodeTask.innerHTML = task;
        return
    }
    this._nodeTask = ScoTools.createHtml('div', null, { 'class': 'auftrag' });
    this._nodeTask.innerHTML = task;
    this._levelMgr.node.parentNode.insertBefore(this._nodeTask, this._levelMgr.node.nextSibling);
}

/**
* Informationen über aktuelles Level ermitteln
* @return Object    Eintrag aus window.roboterCodingTask.levels
*/
RoboterCodingTaskUI.prototype.getLevelInfo = function() {
    const level = window.roboterCodingTask.levels[this._levelMgr.active];
    return level ? level : null;
}

/**
* Level (neu) starten
* @return undefined
*/
RoboterCodingTaskUI.prototype._startLevel = function(
    levelNr,        // Nummer des zu startenden Levels (beginnend bei 0)
    oldLevelNr      // Nummer des vorherigen Levels (beginnend bei 0)
) {
    const level = this.getLevelInfo();
    if (level) {
        if (!this._contentContainer.parentNode) {
            this._levelMgr.node.parentNode.appendChild(this._contentContainer);
        }
        ['_containerQuestion', '_btnNext'].forEach(key => {
            if (this[key]) {
                if (this[key].parentNode)
                    this[key].parentNode.removeChild(this[key]);
                this[key] = null;
            }
        });

        this.clearMsg();
        this._showTask(level.task ? level.task : '');
        if ((typeof level.start == 'function') || (level.start instanceof Function))
            level.start();
        if ((level.code !== undefined) && (!level.keepCode || (oldLevelNr + 1 != levelNr)))
            this._terminal.code = level.code;
        if (level.world)
            this._terminal.world = level.world;
        this._terminal.editingEnabled = !level.readonly;
        this.unlockNext = level.locked ? 0 : 1;
        this.showNextLock();
        /*!*/
    } else {
        if (this._contentContainer.parentNode)
            this._contentContainer.parentNode.removeChild(this._contentContainer);
        this._showTask(this.createFinalReport());
    }
}

/**
* Abschlussbericht zusammenstellen
* @return String        HTML-Code des Abschlussberichts
*/
RoboterCodingTaskUI.prototype.createFinalReport = function() {
    if (!this.fineCount)
        return 'Hervorragend! Du hast alle Aufgaben fehlerfrei gelöst.';
    if (this.fineCount == 1)
        return 'Gut gemacht: Du hast alle Aufgaben gelöst und nur <strong>einen Fehlerpunkt</strong> kassiert.';
    const arrZahlen = ['', '', 'zwei', 'drei', 'vier', 'fünf', 'sechs', 'sieben', 'acht', 'neun', 'zehn', 'elf', 'zwölf'];
    if (this.fineCount < arrZahlen.length)
        return 'Du hast alle Aufgaben gemeistert, dabei allerdings <strong>' + arrZahlen[this.fineCount] + ' Fehlerpunkte</strong> gesammelt.';
    return 'Du hast alle Aufgaben gemeistert, dabei allerdings <strong>' + this.fineCount + ' Fehlerpunkte</strong> gesammelt.';
}

/**
* Levelspezifisch Programmcode vor der Ausführung prüfen
* @return undefined
*/
RoboterCodingTaskUI.prototype._procBeforeExecution = function(
    code,       // Value of this.code
    tokens,     // Array of all tokens representing code
    compiled,   // SourceCodePart parent instance representing parsed code
    worldDef    // Maschine code representing robot's world
) {
    const level = window.roboterCodingTask.levels[this._levelMgr.active];
    const resultRaw = (level && level.checkCode) ? level.checkCode.call(level, {code, tokens, compiled, worldDef}) : undefined;
    const result = (resultRaw === undefined) ? { accept: true } : resultRaw;
    if (result.unlock && (this.unlockNext < 1)) {
        this.unlockNext = 1;
        this.showNextLock();
    }
    if (result.fine) {
        this.fineCount += Number.isFinite(result.fine) ? result.fine : 1;
        /*!*/
    }
    return result;
}

/**
* Levelspezifisch Programmcode nach der Ausführung prüfen
* @return undefined
*/
RoboterCodingTaskUI.prototype._procAfterExecution = function(
    code,       // Value of this.code executed
    line,       // Number of last line executed
    result,     // Reason for termination: undefined if aborted by user, otherwise RobotSimulator.RESULT_*
    worldDef,   // Maschine code representing robot's world before any execution
    worldStat,  // Present robot's world after execution (RobotWorld instance)
    robotStat   // Location and viewing direction of robot after execution: { x, y, direction }
) {
    const level = window.roboterCodingTask.levels[this._levelMgr.active];
    const checkResult = (level && level.checkTermination) ? level.checkTermination.call(level, {code, line, result, worldDef, worldStat, robotX: robotStat.x, robotY: robotStat.y, robotDirection: robotStat.direction }) : null;
    if (checkResult && checkResult.unlock && (this.unlockNext < 1)) {
        this.unlockNext = 1;
        this.showNextLock();
    }
    if (checkResult && checkResult.fine)
        this.fineCount += Number.isFinite(checkResult.fine) ? checkResult.fine : 1;
    if (checkResult && (checkResult.accept === undefined))
        checkResult.accept = true;
    return checkResult;
}

// Anzuzeigende Meldung, wenn es sonst keine Rückmeldung gibt
RoboterCodingTaskUI.prototype._generalMsg = null;

/**
* Festlegen oder löschen, welche generische Meldung anzuzeigen ist, wenn sonst keine Meldung verfügbar ist (und diese gleich anzeigen, wenn eine Meldung festgelegt wird)
* @return undefined
*/
RoboterCodingTaskUI.prototype.setGeneralMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._generalMsg = text ? { text, type } : null;
    if (this._generalMsg)
        this.clearMsg();
}

/**
* Aktive Meldung löschen
* @return undefined
*/
RoboterCodingTaskUI.prototype.clearMsg = function() {
    if (this._generalMsg)
        this._terminal.addMsg(this._generalMsg.text, this._generalMsg.type);
    else
        this._terminal.clearMsg();
}

/**
* Eine Meldung vorübergehend anzeigen
* @return undefined
*/
RoboterCodingTaskUI.prototype.addMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._terminal.addMsg(text, type);
}

/**
* Darauf reagieren, dass Teilaufgabe gelöst wurde
* @return undefined
*/
RoboterCodingTaskUI.prototype.showNextLock = function() {
    if (this.unlockNext < 1)
        return;
    if (this.unlockNext < 2) {
        const level = this.getLevelInfo();
        if (!level || !level.question)
            this.unlockNext = 2;
        else if (!this._containerQuestion) {
            this._containerQuestion = ScoTools.createHtml('form', this._contentContainer, { method: 'post', action: location.href });
            const fragenui = new FragenUI(level.question, this._containerQuestion);
            const submitBtn = ScoTools.createHtml('button', this._containerQuestion, { type: 'submit', class: 'fullwidth' }, 'Antwort überprüfen');
            let needReload = false;
            this._containerQuestion.addEventListener('submit', ev => {
                ev.preventDefault();
                ev.stopPropagation();
                if (needReload) {
                    this._contentContainer.removeChild(this._containerQuestion);
                    this._containerQuestion = null;
                    this.showNextLock();
                    return;
                }
                const result = fragenui.evaluate();
                if (!result.correct) {
                    needReload = true;
                    this.addMsg('Frage falsch beantwortet');
                    submitBtn.textContent = 'Erneut versuchen';
                    if (level.questionWrongFine)
                        this.fineCount += Number.isFinite(level.questionWrongFine) ? level.questionWrongFine : 1;
                    /*!*/// Fehlerpunkte!
                    return;
                }
                submitBtn.parentNode.removeChild(submitBtn);
                if (this.unlockNext < 2)
                    this.unlockNext = 2;
                this.addMsg('Aufgabe gelöst', ScoTools.Message.TYPE_SUCCESS);
                this.showNextLock();
            });
        }
    }
    if (this.unlockNext >= 2) {
        const countDone = this._levelMgr.active + ((this._levelMgr.active + 1 == window.roboterCodingTask.levels.length) ? 2 : 1);
        if (countDone > this._levelMgr.done) {
            this._levelMgr.done = countDone;
            if ((typeof this.onlevelunlock == 'function') || (this.onlevelunlock instanceof Function))
                this.onlevelunlock.call(this, Math.min(countDone, roboterCodingTask.levels.length));
        }
        if (!this._btnNext) {
            this._btnNext = ScoTools.createHtml('button', this._contentContainer, { 'class': 'fullwidth', type: 'button' }, 'Aufgabe abschließen');
            this._btnNext.addEventListener('click', ev => {
                ev.preventDefault();
                ev.stopPropagation();
                ++this._levelMgr.active;
            });
        }
    }
}
